import pygame
from error_handling import CustomError

pygame.init()


class Rect:
    def __init__(self, *args):
        if len(args) == 1:
            args = args[0]

        self.x = args[0]
        self.y = args[1]
        self.width = args[2]
        self.height = args[3]

    def collide(self, *pos):
        if isinstance(pos[0], tuple) or isinstance(pos[0], list):
            x = pos[0][0]
            y = pos[0][1]
        elif len(pos) == 2:
            x = pos[0]
            y = pos[1]
        else:
            return False

        return self.x <= x <= self.x + self.width and self.y <= y <= self.y + self.height

    def update(self, x, y, width, height):
        self.x = x
        self.y = y
        self.width = width
        self.height = height

    def __call__(self):
        return self.x, self.y, self.width, self.height

    def __add__(self, other):
        if isinstance(other, int):
            return self.x - other, self.y - other, self.width + other * 2, self.height + other * 2
        elif isinstance(other, Rect):
            return self.x + other.x, self.y + other.y, self.width + other.width, self.height + other.height

    def __sub__(self, other):
        if isinstance(other, int):
            return self.x + other, self.y + other, self.width - other * 2, self.height - other * 2
        elif isinstance(other, Rect):
            return self.x - other.x, self.y - other.y, self.width - other.width, self.height - other.height

    def __mul__(self, other):
        if isinstance(other, int):
            return self.x - (self.width * (other - 1) / 2), self.y - (self.height * (other - 1) / 2), self.width * other, self.height * other
        elif isinstance(other, Rect):
            return self.x * other.x, self.y * other.y, self.width * other.width, self.height * other.height

    def __truediv__(self, other):
        if isinstance(other, int):
            return self.x + (self.width * (other - 1) / 2), self.y + (self.height * (other - 1) / 2), self.width / other, self.height / other
        elif isinstance(other, Rect):
            return self.x / other.x, self.y / other.y, self.width / other.width, self.height / other.height


class Base:
    def __init__(self, x, y, width, height, fit_text_options=None, base_colour=(255, 255, 255), message="", font_style="arial", font_size=None, font_colour=(0, 0, 0), bold=False, italics=False, antialias=True, alignment="left", border_width=2, border_colour=(0, 0, 0), top_left=0, top_right=0, bottom_left=0, bottom_right=0):
        self.x = x
        self.y = y
        self.width = width
        self.height = height
        # none, text(shrinks the text), width(expands/shrinks the width of the box to fit the text), height(expands/shrinks the height of the box to fit the text), all(both width and height)
        # can be combined in tuple e.g.("text", "height")
        self.fit_text_options = fit_text_options

        self.rect_obj = Rect(self.x, self.y, self.width, self.height)
        self.base_colour = base_colour

        self.message = message
        self.font_style = font_style
        self.font_size = font_size
        self.font_colour = font_colour
        self.bold = bold
        self.italics = italics
        self.antialias = antialias
        self.alignment = alignment

        self.border_width = border_width
        self.border_colour = border_colour

        self.top_left = top_left
        self.top_right = top_right
        self.bottom_left = bottom_left
        self.bottom_right = bottom_right

        self.finish_setup()

    def finish_setup(self):
        if self.font_size is None:
            self.font_size = self.get_font_size(self.message)

        self.fit_text(self.message)

    def fit_text(self, message):
        if not(isinstance(self.fit_text_options, tuple) or isinstance(self.fit_text_options, list)):
            self.fit_text_options = [self.fit_text_options]

        if len(self.fit_text_options) > 1 and self.fit_text_options.__contains__(None):
            raise CustomError("'fit_text_options' field cannot contain 'None' and another value")

        font = pygame.font.SysFont(self.font_style, self.font_size, bold=self.bold, italic=self.italics)
        text = font.render(message, self.antialias, self.font_colour)

        if self.fit_text_options.__contains__("width"):
            self.width = text.get_width()

        if self.fit_text_options.__contains__("height"):
            self.height = text.get_height()

        if self.fit_text_options.__contains__("all"):
            self.width = text.get_width()
            self.height = text.get_height()

        if self.fit_text_options.__contains__("text"):
            self.font_size = self.get_font_size(message)

    def get_font_size(self, message):
        font_size = 50

        if self.fit_text_options is None or self.fit_text_options.__contains__("text"):
            height = 0
            count = 0
            while height != self.height and count < 10:
                count += 1

                font = pygame.font.SysFont(self.font_style, font_size)
                text = font.render("", False, (0, 0, 0))

                height = text.get_height()
                font_size = round(self.height / text.get_height() * font_size)

            if self.fit_text_options is not None:
                font = pygame.font.SysFont(self.font_style, font_size, bold=self.bold, italic=self.italics)
                text = font.render(message, self.antialias, self.font_colour)

                if self.width < text.get_width():
                    font_size = round((self.width - 5) / text.get_width() * font_size)

        return font_size

    def _draw(self, wn):
        self.rect_obj.update(self.x, self.y, self.width, self.height)
        self.fit_text(self.message)

        # border
        pygame.draw.rect(wn, self.border_colour, self.rect_obj + self.border_width, border_top_left_radius=self.top_left, border_top_right_radius=self.top_right, border_bottom_left_radius=self.bottom_left, border_bottom_right_radius=self.bottom_right)
        # box
        pygame.draw.rect(wn, self.base_colour, self.rect_obj(), border_top_left_radius=self.top_left, border_top_right_radius=self.top_right, border_bottom_left_radius=self.bottom_left, border_bottom_right_radius=self.bottom_right)

        # text
        if self.message:
            font = pygame.font.SysFont(self.font_style, self.font_size, bold=self.bold, italic=self.italics)
            text = font.render(self.message, self.antialias, self.font_colour)

            text_pos = (self.x, self.y)
            if self.alignment == "center":
                text_pos = ((self.x + self.width / 2) - (text.get_width() / 2), self.y)
            elif self.alignment == "right":
                text_pos = (self.x + self.width - text.get_width(), self.y)

            wn.blit(text, text_pos)

    def _select(self, mouse_pos):
        self.rect_obj.update(self.x, self.y, self.width, self.height)

        if self.rect_obj.collide(mouse_pos):
            return "box"

        border_rect = Rect(self.rect_obj + self.border_width)
        if border_rect.collide(mouse_pos):
            return "border"

        return False



from text_field_base import *


class Label(Base):
    def __init__(self, x, y, width, height, moveable=False, fit_text_options=None, base_colour=(255, 255, 255), message="", font_style="arial", font_size=None, font_colour=(0, 0, 0), bold=False, italics=False, antialias=True, alignment="left", border_width=2, border_colour=(0, 0, 0), top_left=0, top_right=0, bottom_left=0, bottom_right=0):
        super().__init__(x, y, width, height, fit_text_options, base_colour, message, font_style, font_size, font_colour, bold, italics, antialias, alignment, border_width, border_colour, top_left, top_right, bottom_left, bottom_right)
        self.moveable = moveable
        self.picked_up = False
        self.offset = (0, 0)
        self.point = None
        self.grabbed = False

    def draw(self, wn):
        self._draw(wn)

    def set_mouse(self, mouse_pos):
        if self.moveable:
            if not self.grabbed:
                selected = self._select(mouse_pos)
                if selected == "box":
                    pygame.mouse.set_cursor(pygame.SYSTEM_CURSOR_SIZEALL)
                    self.point = None

                elif selected == "border":
                    if self.x - self.border_width <= mouse_pos[0] <= self.x + self.width * 0.1:
                        if self.y - self.border_width <= mouse_pos[1] <= self.y + self.height * 0.1:
                            pygame.mouse.set_cursor(pygame.SYSTEM_CURSOR_SIZENWSE)
                            self.point = "nw"
                        elif self.y + self.height * 0.9 <= mouse_pos[1] <= self.y + self.height + self.border_width:
                            pygame.mouse.set_cursor(pygame.SYSTEM_CURSOR_SIZENESW)
                            self.point = "sw"
                        else:
                            pygame.mouse.set_cursor(pygame.SYSTEM_CURSOR_SIZEWE)
                            self.point = "w"

                    elif self.x + self.width * 0.9 <= mouse_pos[0] <= self.x + self.width + self.border_width:
                        if self.y - self.border_width <= mouse_pos[1] <= self.y + self.height * 0.1:
                            pygame.mouse.set_cursor(pygame.SYSTEM_CURSOR_SIZENESW)
                            self.point = "ne"
                        elif self.y + self.height * 0.9 <= mouse_pos[1] <= self.y + self.height + self.border_width:
                            pygame.mouse.set_cursor(pygame.SYSTEM_CURSOR_SIZENWSE)
                            self.point = "se"
                        else:
                            pygame.mouse.set_cursor(pygame.SYSTEM_CURSOR_SIZEWE)
                            self.point = "e"
                    else:
                        if self.y - self.border_width <= mouse_pos[1] <= self.y + self.height * 0.1:
                            self.point = "n"
                        elif self.y + self.height * 0.9 <= mouse_pos[1] <= self.y + self.height + self.border_width:
                            self.point = "s"
                        pygame.mouse.set_cursor(pygame.SYSTEM_CURSOR_SIZENS)

                else:
                    pygame.mouse.set_cursor(pygame.SYSTEM_CURSOR_ARROW)
                    self.point = None

        if self.picked_up:
            self.x = mouse_pos[0] + self.offset[0]
            self.y = mouse_pos[1] + self.offset[1]

        if self.grabbed:
            self.resize(mouse_pos)

    def click(self, mouse_pos):
        cursor = pygame.mouse.get_cursor().data[0]
        if cursor == pygame.SYSTEM_CURSOR_SIZEALL:
            self.picked_up = True
            self.offset = (self.x - mouse_pos[0], self.y - mouse_pos[1])
        elif self.point is not None:
            self.grabbed = True

    def unclick(self):
        if self.picked_up:
            self.picked_up = False
        if self.grabbed:
            self.grabbed = False

    def debug(self, *args):
        if args.__contains__("pos"):
            return {"data": (self.x, self.y),
                    "string": f"x: {self.x}\ny: {self.y}"}
        
    def resize(self, mouse_pos):
        if self.point == "nw":
            moved = max(self.x - mouse_pos[0], self.y - mouse_pos[1]) - self.border_width
            self.x -= moved
            self.y -= moved
            self.width += moved
            self.height += moved
        elif self.point == "n":
            moved = self.y - mouse_pos[1] - self.border_width
            self.y -= moved
            self.height += moved
        if self.point == "ne":
            moved = max(mouse_pos[0] - (self.x + self.width), self.y - mouse_pos[1]) - self.border_width
            self.y -= moved
            self.width += moved
            self.height += moved
        if self.point == "e":
            moved = mouse_pos[0] - (self.x + self.width) - self.border_width
            self.width += moved
        if self.point == "se":
            moved = max(mouse_pos[0] - (self.x + self.width), mouse_pos[1] - (self.y + self.height)) - self.border_width
            self.width += moved
            self.height += moved
        if self.point == "s":
            moved = mouse_pos[1] - (self.y + self.height) - self.border_width
            self.height += moved
        if self.point == "sw":
            moved = max(self.x - mouse_pos[0], mouse_pos[1] - (self.y + self.height)) - self.border_width
            self.x -= moved
            self.width += moved
            self.height += moved
        if self.point == "w":
            moved = self.x - mouse_pos[0] - self.border_width
            self.x -= moved
            self.width += moved



from text_field_base import *


class TextBox(Base):
    def __init__(self, x, y, width, height, prompt="", cursor_gap=None, cursor_speed=50, cursor_width=2, cursor_height=None, cursor_colour=(0, 0, 0), fit_text_options=None, base_colour=(255, 255, 255), message="", font_style="arial", font_size=None, font_colour=(0, 0, 0), bold=False, italics=False, antialias=True, alignment="left", border_width=2, border_colour=(0, 0, 0), top_left=0, top_right=0, bottom_left=0, bottom_right=0):
        super().__init__(x, y, width, height, fit_text_options, base_colour, message, font_style, font_size, font_colour, bold, italics, antialias, alignment, border_width, border_colour, top_left, top_right, bottom_left, bottom_right)
        self.prompt = prompt

        self.selected = False
        self.cursor_state = True
        self.cursor_gap = cursor_gap
        self.cursor_speed = cursor_speed

        self.cursor_width = cursor_width
        self.cursor_height = cursor_height
        self.cursor_colour = cursor_colour

        self.cycle = 0

    def draw(self, wn):
        self._draw(wn)

        if self.message == "" and not self.selected:
            self.fit_text(self.prompt)
            font = pygame.font.SysFont(self.font_style, self.font_size, bold=self.bold, italic=self.italics)
            text = font.render(self.prompt, self.antialias, self.font_colour)

            text_pos = (self.x, self.y)
            if self.alignment == "center":
                text_pos = ((self.x + self.width / 2) - (text.get_width() / 2), self.y)
            elif self.alignment == "right":
                text_pos = (self.x + self.width - text.get_width(), self.y)

            wn.blit(text, text_pos)

        if self.selected:
            self.cycle += 1
            if self.cycle % self.cursor_speed == 0:
                self.cursor_state = not self.cursor_state

            if self.cursor_state:
                font = pygame.font.SysFont(self.font_style, self.font_size, bold=self.bold, italic=self.italics)
                text = font.render(self.message, self.antialias, self.font_colour)

                if self.cursor_height is None:
                    cursor_height = text.get_height()
                else:
                    cursor_height = self.cursor_height

                if self.cursor_gap is None:
                    if self.message:
                        cursor_gap = 5 / len(self.message)
                    else:
                        cursor_gap = 5
                else:
                    cursor_gap = self.cursor_gap

                start_pos = (self.x + cursor_gap + text.get_width(), self.y + cursor_gap)
                end_pos = (self.x + cursor_gap + text.get_width(), self.y + cursor_height - cursor_gap)

                pygame.draw.line(wn, self.cursor_colour, start_pos, end_pos, self.cursor_width)

    def click(self, mouse_pos):
        if self._select(mouse_pos):
            self.selected = True
            self.cursor_state = True
            self.cycle = 0
        else:
            self.selected = False

    def key_down(self, key):
        if self.selected:
            if key.key == pygame.K_BACKSPACE:
                self.message = self.message[:-1]
            elif key.key == pygame.K_RETURN or key.key == pygame.K_KP_ENTER:
                self.selected = False
                return True
            else:
                self.message += key.unicode
            return False



def get(filename):
    if not filename.__contains__(".py"):
        filename += ".py"

    with open(filename, "w") as f:
        with open("basic.py", "r") as b:
            f.writelines(b.readlines())
